# -----------------------------------------------------------------------------
#   Copyright (c) 2008 by David P. D. Moss. All rights reserved.
#
#   Released under the BSD license. See the LICENSE file for details.
# -----------------------------------------------------------------------------
"""
The netaddr.contrib namespace for non-core code contributed by users.

It is a testing ground for new ideas. Depending on the interest in
functionality found here, code may find its way into the core in various
ways, either as is or as additions to existing APIs.
"""
